//
//  UserListManager.swift
//  SampleApp
//
//  Created by Nishan Niraula on 5/11/20.
//  Copyright © 2020 David Zhang. All rights reserved.
//

import Foundation
import AmitySDK
import Combine

class UserListManager {
    
    let client: AmityClient
    let userRepository: AmityUserRepository
    let presenceRepository: AmityUserPresenceRepository
    
    var userCollection: AmityCollection<AmityUser>?
    var userCollectionToken: AmityNotificationToken?
    
    var searchCollection: AmityCollection<AmityUser>?
    var searchCollectionToken: AmityNotificationToken?
    
    var isInSearchMode = false
    var searchText = ""
    var sortOption: AmityUserSortOption = .displayName
    
    let debouncer = Debouncer(delay: 0.3)
    var searchedUsers = [AmityUser]()
    
    var cancellables = Set<AnyCancellable>()
    
    init(client: AmityClient) {
        self.client = client
        self.userRepository = AmityUserRepository(client: client)
        self.presenceRepository = AmityUserPresenceRepository(client: client)
    }
    
    @MainActor
    func fetchUserList(sortedBy option: AmityUserSortOption, completion:@escaping ()->()) {
        userCollection = userRepository.getUsers(option)
        userCollectionToken = userCollection?.observe({ (collection, change, error) in
            guard !self.isInSearchMode else { return }

            completion()
        })
        
        // Observe presence state
        presenceRepository.getSyncingUserPresence()
            .sink { completion in
                Log.add(info: "Observer: Completion")
            } receiveValue: { result in
                Log.add(info: "Observer: Synced Results: \(result)")
            }
            .store(in: &cancellables)
    }
    
    func fetchBlockedUserList(completion: @escaping () -> Void) {
        userCollection = userRepository.getBlockedUsers()
        userCollectionToken = userCollection?.observe({ collection, _, error in
            completion()
        })
    }
    
    func searchUserList(name: String, sortBy: AmityUserSortOption, completion:@escaping ()->()) {
        self.searchText = name
        self.sortOption = sortBy
        self.isInSearchMode = !name.isEmpty
        
        if isInSearchMode {
            self.searchedUsers = []
            completion()
            
            debouncer.setCallback { [weak self] in
                self?.waitAndSearchUser(name: name, completion: completion)
            }
            debouncer.call()
        } else {
            completion()
        }
    }
    
    func waitAndSearchUser(name: String, completion:@escaping ()->()) {
        
        searchCollection = userRepository.searchUsers(name, sortBy: self.sortOption, matchType: .partial)
        searchCollectionToken = searchCollection?.observe({ [weak self] (collection, change, error) in
            
            self?.populateSearchedUsers(completion: completion)
        })
    }
    
    func populateSearchedUsers(completion:@escaping ()->()) {
        
        var results = [AmityUser]()
        for item in searchCollection!.allObjects() {
            results.append(item)
        }
        
        self.searchedUsers = results
        
        // Notify tableview to reload
        completion()
    }
    
    func numberOfUsers() -> Int {
        if isInSearchMode {
            return searchedUsers.count
        } else {
            return Int(userCollection?.count() ?? 0)
        }
    }
    
    func getUserItem(at index: Int) -> AmityUser? {
        if isInSearchMode, index < searchedUsers.count {
            guard !searchedUsers.isEmpty else { return nil }
            
            return searchedUsers[index]
        } else {
            return userCollection?.object(at: index)
        }
    }
    
    func loadMoreUsers() {
        if isInSearchMode {
            guard let hasMorePosts = searchCollection?.hasNext, hasMorePosts else { return }
            searchCollection?.nextPage()
        } else {
            guard let hasMorePosts = userCollection?.hasNext, hasMorePosts else { return }
            userCollection?.nextPage()
        }
    }
}
